#!/usr/bin/env python


import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################
# -------------------------------------------------------------------------------------


# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>


#from gimpfu import *
#from array import array
#import math
#import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()


#####
##### TEMP VALUES

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
darkestgrey = gimpcolor.RGB(20, 20, 20)
verydarkgrey = gimpcolor.RGB(50, 50, 50)
darkgrey = gimpcolor.RGB(75, 75, 75)
midgrey2 = gimpcolor.RGB(105, 105, 105)
midgrey = gimpcolor.RGB(125, 125, 125)
lightgrey = gimpcolor.RGB(170, 170, 170)
verylightgrey = gimpcolor.RGB(222, 222, 222)
red = gimpcolor.RGB(220,10,10)	





def do_stuff(image, layer, direction, main_blur_amount, main_bump_depth, main_elevation, do_preview, preview_char,
            do_edge_highlight, do_edge_shadow, edge_hi_blur, edge_hi_bump_depth, edge_hi_elevation, edge_smoothing, edge_highlight_opacity,  
            do_extra, intensify, shrinkage, highlight_bump_depth, highlight_elevation, highlight_colour, highlight_blur, hi_opacity, hi_mode, 
            do_bloom, bloom_expansion, highlight_smoothing, bloom_opacity, bloom_mode, bloom_colour, distort, fairy_dust, ptab):


    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
	
    #####
    ##### TEMP VALUES
    apply_to_custom = True
	
	
    fg_col = pdb.gimp_context_get_foreground()
    bg_col = pdb.gimp_context_get_background()
    original_fg_col = pdb.gimp_context_get_foreground()
    original_bg_col = pdb.gimp_context_get_background()
		
    dummy = 0
    active_layer_id = pdb.gimp_image_get_active_layer(image)
	
	
    loop_ptr = 1
    first_char_layer_name = "none"
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers
    original_active_layer = image.active_layer
	
	
    if preview_char == "": preview_char = "A"
    preview_char = preview_char + ".png"
	
    # check that the preview character is present or use first available
    if preview_char in str(layer_names_list):
        prev_pos = str(layer_names_list).find(preview_char)
        prev_loop = str(layer_names_list).count(",",0,prev_pos)
    else:
        loop = 0
        preview_char = pdb.gimp_item_get_name(layer_names_list[loop])
        while preview_char=="space.png" or (preview_char not in str(layer_names_list[loop])):
            loop += 1
            preview_char = pdb.gimp_item_get_name(layer_names_list[loop])

    preview_char = preview_char.decode('utf-8')	######## Added to catch higher value unicodes
	
	
    if do_preview:
        # copy the preview character and paste it as a  new image
		
        preview_layer = pdb.gimp_image_get_layer_by_name(image, preview_char)
        success = pdb.gimp_edit_copy(preview_layer)
        preview_image = pdb.gimp_edit_paste_as_new_image(preview_layer)
            
        # Set end undo point for preview image
        pdb.gimp_image_undo_group_start(preview_image) ##############
		
        preview_display = pdb.gimp_display_new(preview_image)

        num_preview_layers, preview_layer_ids = pdb.gimp_image_get_layers(preview_image)
        preview_layer_names_list = preview_image.layers
		
        loop = 0
        layer_id = preview_layer_ids[loop]
        layer_name = preview_layer_names_list[loop]
		
        # cater for unicode characters above normal range
        display_layer_name =  pdb.gimp_item_get_name(preview_layer_names_list[loop])
        display_layer_name = display_layer_name.decode('utf-8')
	
        # check we have the layer id of the currently active layer
        pdb.gimp_image_set_active_layer(preview_image, layer_name)
        active_layer_id = pdb.gimp_image_get_active_layer(preview_image)
		
        active_layer = preview_image.active_layer
        active_layer_id = proc_do_highlights(preview_image, active_layer, original_fg_col, original_bg_col, direction, main_bump_depth, main_elevation, main_blur_amount, do_preview, preview_char,
            do_edge_highlight, do_edge_shadow,edge_hi_blur, edge_hi_bump_depth, edge_hi_elevation, edge_smoothing, edge_highlight_opacity, 
            do_extra, intensify, shrinkage, highlight_bump_depth, highlight_elevation, highlight_colour, highlight_blur, hi_mode, hi_opacity, 
            do_bloom, bloom_expansion, highlight_smoothing, bloom_opacity, bloom_mode, bloom_colour, distort, fairy_dust)
        pdb.gimp_item_set_visible(active_layer_id, False)
		
        pdb.gimp_item_set_visible(active_layer_id, True)
		
        pdb.gimp_displays_flush()
		
        # Set end undo point for preview image
        pdb.gimp_image_undo_group_end(preview_image)
		
    else:
	
		
        # Set start undo point
        pdb.gimp_image_undo_group_start(image)	
        num_multi_layers,other_layers,num_char_layers = 0,0,0
        visi_loop, first_char_layer  = 0, -1

        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
        while visi_loop < num_layers:
	        # get info about layer we are working on
            layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            layer_name = layer_name.decode('utf-8')
            # check to see if the current layer is a custom font character layer
            test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[visi_loop], layer_name)
			
            if test_for_custom_layer:
                num_char_layers += 1
                if first_char_layer < 0: 
                    first_char_layer = visi_loop
                    first_char_layer_name = layer_name
                    layer_names_list[visi_loop].set_offsets(0, 0)
				
            
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) # turn off visibility
            visi_loop += 1
		
		
        while loop >= 0:
	        # get info about layer we are working on for custom font mode
            layer_id = layer_ids[loop]
            layer_name = layer_names_list[loop]
	
            # cater for unicode characters above normal range
            display_layer_name =  pdb.gimp_item_get_name(layer_names_list[loop])
            display_layer_name = display_layer_name.decode('utf-8')	
	
            # check we have the layer id of the currently active layer
            pdb.gimp_image_set_active_layer(image, layer_name)
            layer_id = pdb.gimp_image_get_active_layer(image)

            # check to see if the current layer is a custom font character layer
            test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, display_layer_name)
	
	
            if test_for_custom_layer:
                # ensure we have the active layer and make it visible
                active_layer_id = pdb.gimp_image_get_active_layer(image)
                pdb.gimp_item_set_visible(active_layer_id, True)
			
                active_layer = image.active_layer
                active_layer_id = proc_do_highlights(image, active_layer, original_fg_col, original_bg_col, direction, main_bump_depth, main_elevation, main_blur_amount, do_preview, preview_char,
                    do_edge_highlight, do_edge_shadow,edge_hi_blur, edge_hi_bump_depth, edge_hi_elevation, edge_smoothing, edge_highlight_opacity, 
                    do_extra, intensify, shrinkage, highlight_bump_depth, highlight_elevation, highlight_colour, highlight_blur, hi_mode, hi_opacity, 
                    do_bloom, bloom_expansion, highlight_smoothing, bloom_opacity, bloom_mode, bloom_colour, distort, fairy_dust)
                pdb.gimp_item_set_visible(active_layer_id, False)
			
            loop_ptr += 1
            loop = num_layers - loop_ptr
		
		
        if first_char_layer_name <> "none": # the character layer to make visible on completion
            first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
            pdb.gimp_item_set_visible(first_char_layer_id, True)
            pdb.gimp_image_set_active_layer(image,first_char_layer_id)
	
        pdb.gimp_displays_flush()
		
        # Set end undo point
        pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

    #### Useful Functions
# -------------------------------------------------------------------------------------
def proc_do_highlights(image, layer, original_fg_col, original_bg_col, direction, main_bump_depth, main_elevation, main_blur_amount, do_preview, preview_char,
                do_edge_highlight, do_edge_shadow,edge_hi_blur, edge_hi_bump_depth, edge_hi_elevation, edge_smoothing, edge_highlight_opacity, 
                do_extra, intensify, shrinkage, highlight_bump_depth, highlight_elevation, highlight_colour, highlight_blur, hi_mode, hi_opacity, 
                do_bloom, bloom_expansion, highlight_smoothing, bloom_opacity, bloom_mode, bloom_colour, distort, fairy_dust):
	
    main_layer = layer
	
    # add initial bumpmap layer to character
    bump_layer=pdb.gimp_layer_new(image, main_layer.width, main_layer.height, 1, "bump layer", 100, 28)
    pdb.gimp_image_insert_layer(image, bump_layer, None, -1 )
    pdb.gimp_image_lower_item(image,bump_layer)
    bump_bg = pdb.gimp_layer_copy(bump_layer, True)
    pdb.gimp_image_insert_layer(image, bump_bg, None, -1 )
    pdb.gimp_image_lower_item(image,bump_bg)
    pdb.gimp_item_set_name(bump_bg,"bump bg")
    pdb.gimp_context_set_foreground(black)
    pdb.gimp_drawable_edit_fill(bump_bg, 0)
    pdb.gimp_image_select_item(image, 2, main_layer)
    pdb.gimp_context_set_foreground(white)
    pdb.gimp_drawable_edit_fill(bump_layer, 0)
    pdb.gimp_selection_none(image)
    pdb.plug_in_gauss_rle2(image, bump_layer, main_blur_amount, main_blur_amount)
    bump_bg = pdb.gimp_image_merge_down(image, bump_layer, 0)
    pdb.plug_in_bump_map(image, main_layer, bump_bg, direction, main_elevation, main_bump_depth, 0, 0, 0, 0, True, False, 0)
    pdb.plug_in_sel_gauss(image, main_layer, 10.0, 100)
    pdb.gimp_item_set_visible(bump_bg, False)
	
    # add a highlight/shadow with a second bumpmap to the edges of the character if required
    if do_edge_highlight or do_edge_shadow:
        pdb.gimp_image_set_active_layer(image, main_layer)
        highlight_layer = pdb.gimp_layer_copy(main_layer, True)
        pdb.gimp_image_insert_layer(image, highlight_layer, None, -1 )
        pdb.gimp_item_set_name(highlight_layer,"highlight layer (highlight_layer)")
        pdb.gimp_drawable_edit_clear(highlight_layer)
	
	
        pdb.gimp_image_select_item(image, 2, main_layer)
        pdb.gimp_context_set_foreground(midgrey)
        pdb.gimp_drawable_edit_fill(highlight_layer, 0)
		
        highlight_channel = pdb.gimp_selection_save(image)
        pdb.gimp_item_set_name(highlight_channel,"highlight channel")
        pdb.gimp_selection_none(image)
        pdb.plug_in_gauss_rle2(image, highlight_channel, (500/100)*edge_hi_blur, (500/100)*edge_hi_blur)
        pdb.gimp_image_set_active_layer(image, highlight_layer)
        pdb.plug_in_bump_map(image, highlight_layer, highlight_channel, direction, edge_hi_elevation, edge_hi_bump_depth, 0, 0, 0, 0, True, False, 0)
        pdb.plug_in_sel_gauss(image, highlight_layer, 10.0, 100)
        pdb.plug_in_colortoalpha(image, highlight_layer, midgrey)
        if not do_edge_shadow: pdb.plug_in_colortoalpha(image, highlight_layer, black)
        if not do_edge_highlight: pdb.plug_in_colortoalpha(image, highlight_layer, white)
        #if do_edge_highlight or do_edge_shadow: 
		
        pdb.gimp_image_select_item(image, 2, main_layer)
        pdb.plug_in_gauss_rle2(image, highlight_layer, edge_smoothing, edge_smoothing)
        pdb.gimp_selection_none(image)
        pdb.gimp_layer_set_opacity(highlight_layer, edge_highlight_opacity)
        
        main_layer = pdb.gimp_image_merge_down(image,highlight_layer,0)
        pdb.gimp_image_remove_channel(image, highlight_channel)
    pdb.gimp_image_remove_layer(image, bump_bg)
	
    if do_extra:
        # add a second highlight - the main highlight
        pdb.gimp_image_set_active_layer(image, main_layer)
        extra_layer = pdb.gimp_layer_copy(main_layer, True)
        pdb.gimp_image_insert_layer(image, extra_layer, None, -1 )
        pdb.gimp_item_set_name(extra_layer,"extra layer (extra_layer)")
        pdb.gimp_drawable_edit_clear(extra_layer)
		
		
        pdb.gimp_image_select_item(image, 2, main_layer)
        pdb.gimp_context_set_foreground(midgrey)
        pdb.gimp_drawable_edit_fill(extra_layer, 0)
        extra_channel = pdb.gimp_selection_save(image)
        pdb.gimp_item_set_name(extra_channel,"extra channel")
        pdb.gimp_selection_none(image)
        pdb.plug_in_gauss_rle2(image, extra_channel, edge_hi_blur, edge_hi_blur) # REM add option for this blur
		
		
        pdb.gimp_image_select_item(image, 2, extra_channel)
        pdb.gimp_selection_shrink(image, shrinkage)
        small_highlight_channel = pdb.gimp_selection_save(image)
        pdb.gimp_item_set_name(small_highlight_channel,"small highlight channel")
        pdb.gimp_selection_none(image)
        small_highlight_layer = pdb.gimp_layer_copy(main_layer, True)
        pdb.gimp_image_set_active_layer(image, main_layer)
        pdb.gimp_image_insert_layer(image, small_highlight_layer, None, -1 )
        pdb.gimp_item_set_name(small_highlight_layer,"small highlight layer (small_highlight_layer)")
        pdb.gimp_drawable_edit_clear(small_highlight_layer)
        pdb.gimp_image_select_item(image, 2, small_highlight_channel)
        #empty = pdb.gimp_selection_is_empty(image)
        pdb.gimp_drawable_edit_fill(small_highlight_layer, 0)
        pdb.gimp_selection_none(image)
        #pdb.plug_in_gauss_rle2(image, small_highlight_channel, 10, 10)
        pdb.gimp_image_set_active_layer(image, small_highlight_layer)
        pdb.plug_in_bump_map(image, small_highlight_layer, small_highlight_channel, direction, highlight_bump_depth, highlight_elevation, 0, 0, 0, 0, True, False, 0)
        pdb.gimp_image_raise_item(image, small_highlight_layer)
        pdb.plug_in_colortoalpha(image, small_highlight_layer, midgrey)
        pdb.plug_in_colortoalpha(image, small_highlight_layer, black)
		
        if highlight_colour>0: # fill with current foreground colour if required
            pdb.gimp_image_select_item(image, 2, small_highlight_layer)
            empty = pdb.gimp_selection_is_empty(image)
            if not empty: # small thin characters may not give a selection
			
                pdb.gimp_context_set_foreground(original_fg_col)
                pdb.gimp_drawable_edit_fill(small_highlight_layer, 0)
            pdb.gimp_selection_none(image)
        pdb.plug_in_gauss_rle2(image, small_highlight_layer, highlight_blur, highlight_blur)
        if intensify: # attempt to emphasize highlight 
            intense_pos = pdb.gimp_image_get_item_position(image, small_highlight_layer)
            intense_layer = pdb.gimp_layer_copy(small_highlight_layer, True)
            pdb.gimp_image_insert_layer(image, intense_layer, None, intense_pos )
            small_highlight_layer  = pdb.gimp_image_merge_down(image,intense_layer,0)
            
        new_hi_mode=fn_get_mode(hi_mode)
        pdb.gimp_layer_set_mode(small_highlight_layer,new_hi_mode)
        pdb.gimp_layer_set_opacity(small_highlight_layer, hi_opacity)
		
        extra_pos = pdb.gimp_image_get_item_position(image, extra_layer)
        pdb.gimp_drawable_edit_clear(extra_layer)
		
        if do_bloom: # add a haloed area 
            pdb.gimp_image_select_item(image, 2, main_layer)
            if bloom_expansion>0: pdb.gimp_selection_shrink(image, bloom_expansion)
            is_empty = pdb.gimp_selection_is_empty(image)
            if is_empty:pdb.gimp_message("Unable to add bloom - increase highlight bloom area value.")
            if not is_empty:
                if distort:
                    random.seed()
                    threshold = random.randrange(80,201,10)
                    spread = random.randrange(4,25,1)
                    smooth = random.randrange(1,4,1)
                    pdb.script_fu_distress_selection(image, small_highlight_layer, threshold, spread, 4, smooth, True, True)
				
            
				
                fg_col = white
                if bloom_colour==1:
                   fg_col = original_fg_col
                elif bloom_colour==2:
                   fg_col = original_bg_col
                pdb.gimp_context_set_foreground(fg_col)
                pdb.gimp_drawable_edit_fill(extra_layer, 0)
                pdb.gimp_selection_none(image)
                if highlight_smoothing>0: pdb.plug_in_gauss_rle2(image, extra_layer, (500/100)*highlight_smoothing, (500/100)*highlight_smoothing) # REM add option for this blur
		
                if fairy_dust > 0:
                    pdb.gimp_image_select_item(image, 2, extra_layer)
                    hsv_value = fairy_dust * 50
                    #pdb.plug_in_hsv_noise(image, extra_layer, 1,50, 50, hsv_value)
                    pdb.plug_in_hsv_noise(image, main_layer, 1,0, hsv_value, 0)
                    pdb.gimp_selection_none(image)
		
                pdb.gimp_image_select_item(image, 2, main_layer)
                pdb.gimp_selection_invert(image)
                pdb.gimp_drawable_edit_clear(extra_layer)
                pdb.gimp_selection_none(image)
		
                new_bloom_mode=fn_get_mode(bloom_mode)
                pdb.gimp_layer_set_mode(extra_layer,new_bloom_mode)
                pdb.gimp_layer_set_opacity(extra_layer, bloom_opacity)
			
        highlight_layer = pdb.gimp_image_merge_down(image,extra_layer,0)
        if do_edge_highlight or do_edge_shadow:main_layer = pdb.gimp_image_merge_down(image,small_highlight_layer,0)
        if not do_edge_highlight and not do_edge_shadow:main_layer = pdb.gimp_image_merge_down(image,small_highlight_layer,0)
        pdb.gimp_image_remove_channel(image, extra_channel)
        pdb.gimp_image_remove_channel(image, small_highlight_channel)
        pdb.gimp_displays_flush()
    return main_layer
# -------------------------------------------------------------------------------------
######################################################################################################################################

def fn_get_mode(new_mode):
        temp_mode = 0
        if new_mode <> 0:
            if new_mode == 1:
                temp_mode = 1
            elif new_mode == 2:
                temp_mode = 57
            elif new_mode == 3:
                temp_mode = 59
            elif new_mode == 4:
                temp_mode = 60
            elif new_mode == 5:
                temp_mode = 36
            elif new_mode == 6:
                temp_mode = 55
            elif new_mode == 7:
                temp_mode = 31
            elif new_mode == 8:
                temp_mode = 42
            elif new_mode == 9:
                temp_mode = 33
            elif new_mode == 10:
                temp_mode = 35
            elif new_mode == 11:
                temp_mode = 54
            elif new_mode == 12:
                temp_mode = 30
            elif new_mode == 13:
                temp_mode = 43
            elif new_mode == 14:
                temp_mode = 53
            elif new_mode == 15:
                temp_mode = 23
            elif new_mode == 16:
                temp_mode = 45
            elif new_mode == 17:
                temp_mode = 44
            elif new_mode == 18:
                temp_mode = 48
            elif new_mode == 19:
                temp_mode = 49
            elif new_mode == 20:
                temp_mode = 50
            elif new_mode == 21:
                temp_mode = 51
            elif new_mode == 22:
                temp_mode = 32
            elif new_mode == 23:
                temp_mode = 52
            elif new_mode == 24:
                temp_mode = 34
            elif new_mode == 25:
                temp_mode = 46
            elif new_mode == 26:
                temp_mode = 47
            elif new_mode == 27:
                temp_mode = 41
            elif new_mode == 28:
                temp_mode = 37
            elif new_mode == 29:
                temp_mode = 38
            elif new_mode == 30:
                temp_mode = 39
            elif new_mode == 31:
                temp_mode = 40
            elif new_mode == 32:
                temp_mode = 24
            elif new_mode == 33:
                temp_mode = 25
            elif new_mode == 34:
                temp_mode = 26
            elif new_mode == 35:
                temp_mode = 27
            elif new_mode == 36:
                temp_mode = 56
        mode = temp_mode
			

        return mode
	

######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Highlights_v1_0_Plug_in_Tabbed",
    "Provides options for adding highlights to custom fonts",
    "Provides options for adding highlights to custom fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2023",
    "<Image>/Filters/Custom Fonts/Decorating Tools/Custom Font Highlights v0.1 Tabbed...",
    "RGB*",
    [
    (PF_SPINNER, "Light_Direction", "Light Direction (Degrees): ", 135, (0, 360, 45),{'newLine':'1','tab':0}),
    (PF_SPINNER, "scaled_scale", "Main bump-map amount: ", 20, (5,30,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "bump_depth", "Main bump-map depth: ", 10, (5,64,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "elevation", "Main bump-map light elevation: ", 35, (15,45,1),{'newLine':'1','tab':0}),
    (PF_TOGGLE, "Preview_option", "Preview single character\n(to a new image): ", 0, [("No Preview"),("Preview") ],{'newLine':'1','tab':0}),
    (PF_STRING, "preview_string", "Preview character: ", "A",{'newLine':'1','tab':0}),
	
    (PF_TOGGLE, "doedgeHighlight", "Add highlight to edge: ", 0, [("No edge highlight"),("Add edge highlight") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "doedgeShadow", "Add shadow to edge: ", 0, [("No edge shadow"),("Add edge shadow") ],{'newLine':'1','tab':1}),
    (PF_SPINNER, "edgehigh_blur", "Edge highlight/shadow bump-map amount: ", 5, (5,30,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "edgehigh_bump_depth", "Edge highlight/shadow bump-map depth: ", 5, (5,64,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "edgehigh_elevation", "Edge highlight/shadow elevation: ", 35, (5,45,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "edgehigh_smoothing", "Edge highlight/shadow smoothing: ", 5, (1,50,1),{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_edgehigh_opacity", "Edge highlight/shadow opacity: ", 100, (5, 100, 1),{'newLine':'1','tab':1}),
	
    (PF_TOGGLE, "doextra", "Add a first highlight: ", 0, [("No first highlight"),("Add first highlight") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "dointense", "Intensify first highlight: ", 0, [("Normal highlight"),("Intensify highlight") ],{'newLine':'1','tab':2}),
    (PF_SPINNER, "shrink_value", "First highlight offset: ", 5, (1,100,1),{'newLine':'1','tab':2}),
    (PF_SPINNER, "high_bump_depth", "First highlight bump-map depth: ", 10, (1,64,1),{'newLine':'1','tab':2}),
    (PF_SPINNER, "high_elevation", "First highlight elevation: ", 25, (5,45,1),{'newLine':'1','tab':2}),
    (PF_OPTION, "extra_colours", "Colour for first highlight: ", 0, ("White", "Foreground colour"),{'newLine':'1','tab':2}),
    (PF_SPINNER, "high_blurring", "First highlight blur: ", 5, (1,50,1),{'newLine':'1','tab':2}),
    (PF_SLIDER, "pf_high_opacity", "First highlight opacity: ", 100, (5, 100, 1),{'newLine':'1','tab':2}),
    (PF_OPTION, "mode_option1", "Mode to use with first highlight: ", 0, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", 
                "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", 
                "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance"),{'newLine':'1','tab':2}),
 
    (PF_TOGGLE, "dobloom", "Add second highlight: ", 0, [("No second highlight"),("Add second highlight") ],{'newLine':'1','tab':3}),
    (PF_SPINNER, "high_expansion", "Reduce second highlight area: ", 5, (0,100,1),{'newLine':'1','tab':3}),
    (PF_SPINNER, "high_smoothing", "Second highlight blur: ", 5, (0,50,1),{'newLine':'1','tab':3}),
    (PF_SLIDER, "pf_bloom_opacity", "Second highlight opacity: ", 20, (5, 100, 1),{'newLine':'1','tab':3}),
    (PF_OPTION, "mode_option2", "Mode to use with second highlight: ", 26, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", 
                "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", 
                "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance"),{'newLine':'1','tab':3}),
    (PF_OPTION, "bloom_colours", "Colour for second highlight: ", 0, ("White", "Foreground colour", "Background colour"),{'newLine':'1','tab':3}),
    (PF_TOGGLE, "dist_bloom", "Distort the second highlight: ", 0, [("No distortion"),("Distort highlight") ],{'newLine':'1','tab':3}),
    (PF_SPINNER, "dust", "Spoonfuls of fairy dust: ", 0, (0,5,1),{'newLine':'1','tab':3}),
	
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Main character'),_('Edge'),_('First highlight'),_('Second highlight')) ),
    ],
    [],
    do_stuff)

main()
